@@ -2,8 +2,10 @@ import os
 import sys
 import contextlib
 from io import StringIO
-import nbformat
+import itertools
 
+import parso
+import nbformat
 from IPython.core.interactiveshell import InteractiveShell
 from IPython.core.displaypub import DisplayPublisher
 from IPython.core.displayhook import DisplayHook
@@ -74,6 +76,11 @@ class PloomberShell(InteractiveShell):
     """
     A subclass of IPython's InteractiveShell to gather all the output
     produced by a code cell
+
+    Notes
+    -----
+    This is intended to be used as a singleton, so either call
+    `.clear_instance()` when you're done or use it as a context manager
     """
 
     def __init__(self):
@@ -112,6 +119,12 @@ class PloomberShell(InteractiveShell):
         configure_inline_support.current_backend = 'unset'
         return super().enable_matplotlib(gui)
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self.clear_instance()
+
 
 class PloomberClient:
     """
@@ -123,6 +136,11 @@ class PloomberClient:
         self._nb = nb
         self._shell = None
 
+    @classmethod
+    def from_path(cls, path):
+        nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
+        return cls(nb)
+
     def execute_cell(self, cell, cell_index, execution_count, store_history):
         if self._shell is None:
             raise RuntimeError('A shell has not been initialized')
@@ -131,7 +149,7 @@ class PloomberClient:
         # and return them
         with patch_sys_std_out_err() as (stdout_stream, stderr_stream):
             result = self._shell.run_cell(cell['source'])
-            stdout = stdout_stream.getvalue()
+            stdout = stdout_stream.get_separated_values()
             stderr = stderr_stream.getvalue()
 
         output = []
@@ -165,13 +183,48 @@ class PloomberClient:
     def execute(self):
         """Execute the notebook
         """
-        # make sure that the current working directory is in the sys.path
-        # in case the user has local modules
         with self:
             self._execute()
 
         return self._nb
 
+    def get_namespace(self):
+        """Run the notebook and return all the variables
+        """
+        with self:
+            existing = set(self._shell.user_ns)
+            self._execute()
+            namespace = {
+                k: v
+                for k, v in self._shell.user_ns.items() if k not in existing
+            }
+
+        return namespace
+
+    def get_definitions(self):
+        """Returns class and function definitions without running the notebook
+        """
+        source = '\n'.join(
+            [c.source for c in self._nb.cells if c.cell_type == 'code'])
+        module = parso.parse(source)
+
+        defs = [
+            def_.get_code() for def_ in itertools.chain(
+                module.iter_classdefs(), module.iter_funcdefs())
+        ]
+
+        defs_source = '\n'.join(defs)
+
+        with self:
+            existing = set(self._shell.user_ns)
+            self._shell.run_cell(defs_source)
+            defs = {
+                k: v
+                for k, v in self._shell.user_ns.items() if k not in existing
+            }
+
+        return defs
+
     def _execute(self):
         """
         Internal method to execute a notebook, assumes the shell has been
@@ -179,6 +232,8 @@ class PloomberClient:
         """
         execution_count = 1
 
+        # make sure that the current working directory is in the sys.path
+        # in case the user has local modules
         with add_to_sys_path('.'):
             for index, cell in enumerate(self._nb.cells):
                 if cell.cell_type == 'code':
@@ -239,15 +294,17 @@ class PloomberManagedClient(PloomberClient):
         return self._nb
 
 
-class IO:
+class IO(StringIO):
 
-    def __init__(self):
+    def __init__(self, initial_value='', newline='\n'):
+        super().__init__(initial_value=initial_value, newline=newline)
         self._values = []
 
     def write(self, s):
         self._values.append(s)
+        super().write(s)
 
-    def getvalue(self):
+    def get_separated_values(self):
         return self._values[:]
 
 
